"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const fs_extra_1 = __importDefault(require("fs-extra"));
const ramda_1 = require("ramda");
const path_1 = __importDefault(require("path"));
const typescript_1 = __importDefault(require("typescript"));
const crypto_1 = __importDefault(require("crypto"));
const cache = { fileNameMapping: {} };
function getTSConfigPath(cwd) {
    const configPath = typescript_1.default.findConfigFile(cwd, typescript_1.default.sys.fileExists, 'tsconfig.json');
    return configPath;
}
function getTSConfig(cwd) {
    const tsconfigPath = getTSConfigPath(cwd);
    if (!tsconfigPath) {
        throw new Error("Could not find a valid 'tsconfig.json'.");
    }
    const tsconfig = require(tsconfigPath);
    return tsconfig;
}
const parseConfigHost = {
    fileExists: fs_extra_1.default.existsSync,
    readDirectory: typescript_1.default.sys.readDirectory,
    readFile: function (file) {
        return fs_extra_1.default.readFileSync(file, 'utf8');
    },
    useCaseSensitiveFileNames: true,
};
function getFileNames(cwd) {
    const tsconfigPath = getTSConfigPath(cwd);
    const tsconfig = getTSConfig(cwd);
    if (tsconfigPath) {
        const parsed = typescript_1.default.parseJsonConfigFileContent(tsconfig, parseConfigHost, path_1.default.dirname(tsconfigPath));
        return parsed.fileNames;
    }
    return [];
}
const files = {};
function getTSService(options, cwd) {
    if (cache.languageService) {
        return cache.languageService;
    }
    const rootFileNames = getFileNames(cwd);
    rootFileNames.forEach((fileName) => {
        files[fileName] = { version: 0, hashedContent: hash(fs_extra_1.default.readFileSync(fileName).toString()) };
    });
    const servicesHost = {
        getScriptFileNames: () => rootFileNames,
        getScriptVersion: (fileName) => files[fileName] && files[fileName].version.toString(),
        getScriptSnapshot: (fileName) => {
            if (!fs_extra_1.default.existsSync(fileName)) {
                return undefined;
            }
            return typescript_1.default.ScriptSnapshot.fromString(fs_extra_1.default.readFileSync(fileName).toString());
        },
        getCurrentDirectory: () => cwd,
        getCompilationSettings: () => options,
        getDefaultLibFileName: (options) => typescript_1.default.getDefaultLibFilePath(options),
        fileExists: typescript_1.default.sys.fileExists,
        readFile: typescript_1.default.sys.readFile,
        readDirectory: typescript_1.default.sys.readDirectory,
        directoryExists: typescript_1.default.sys.directoryExists,
        getDirectories: typescript_1.default.sys.getDirectories,
    };
    const languageService = typescript_1.default.createLanguageService(servicesHost, typescript_1.default.createDocumentRegistry());
    cache.languageService = languageService;
    return languageService;
}
function normalizePath(fileName) {
    if (path_1.default.sep !== path_1.default.posix.sep) {
        return fileName.split(path_1.default.sep).join(path_1.default.posix.sep);
    }
    return fileName;
}
function emitFile(context, languageService, loaderOptions) {
    const fileName = context.resourcePath;
    try {
        const output = languageService.getEmitOutput(fileName);
        if (!output.emitSkipped) {
            output.outputFiles.forEach((o) => {
                if (o.name.endsWith('.d.ts')) {
                    fs_extra_1.default.ensureDirSync(path_1.default.dirname(o.name));
                    fs_extra_1.default.writeFileSync(o.name, o.text);
                    if (loaderOptions.exposes &&
                        !(0, ramda_1.isEmpty)(loaderOptions.exposes) &&
                        !(0, ramda_1.isEmpty)(loaderOptions.name)) {
                        for (const [key, value] of Object.entries(loaderOptions.exposes)) {
                            if (key && value) {
                                context.resolve(context.rootContext, value, (err, inputFilePath) => {
                                    if (err) {
                                        console.error(err);
                                        return;
                                    }
                                    if (inputFilePath === fileName) {
                                        const moduleFilename = `${key}.d.ts`;
                                        const modulePath = path_1.default.resolve(context.rootContext, `${loaderOptions.typesOutputDir}/${loaderOptions.name}`);
                                        const dtsEntryPath = path_1.default.resolve(modulePath, moduleFilename);
                                        const relativePathToOutput = normalizePath(path_1.default.relative(path_1.default.dirname(dtsEntryPath), o.name.replace('.d.ts', '')));
                                        fs_extra_1.default.ensureFileSync(dtsEntryPath);
                                        fs_extra_1.default.writeFileSync(dtsEntryPath, `export * from './${relativePathToOutput}';\nexport { default } from './${relativePathToOutput}';`);
                                    }
                                });
                            }
                        }
                    }
                }
            });
        }
    }
    catch (e) {
        console.log(`Skip ${fileName}`);
    }
}
function hash(content) {
    const hash = crypto_1.default.createHash("md5");
    hash.update(content);
    return hash.digest('hex');
}
function makeLoader(context, loaderOptions, content) {
    const fileName = normalizePath(context.resourcePath);
    const hashedContent = hash(content);
    if (files[fileName] && files[fileName].hashedContent !== hashedContent) {
        files[fileName].hashedContent = hashedContent;
        files[fileName].version++;
    }
    const tsconfig = getTSConfig(context.rootContext);
    const languageService = getTSService({
        ...tsconfig,
        declaration: true,
        emitDeclarationOnly: true,
        outDir: path_1.default.resolve(context.rootContext, `${loaderOptions.typesOutputDir}/${loaderOptions.name}/dts`),
    }, context.rootContext);
    emitFile(context, languageService, loaderOptions);
    return content;
}
function loader(content) {
    const context = this;
    const loaderOptions = context.getOptions();
    return makeLoader(context, {
        name: loaderOptions.name,
        exposes: loaderOptions.exposes,
        typesOutputDir: loaderOptions.typesOutputDir || '.wp_federation',
    }, content);
}
exports.default = loader;
